# -*- coding: utf-8 -*-
"""
Overview
=======
 Handles interactions with a Quickshot server.

Legal
=====
 All code, unless otherwise indicated, is original, and subject to the terms of
 the attached licensing agreement.
 
 Copyright (c) 2010
  Neil Tallim <flan@uguu.ca>
 
 Quickshot is free software: you can redistribute it and/or modify it 
 under the terms of the GNU General Public License version 3, as published 
 by the Free Software Foundation.
 
 Quickshot is distributed in the hope that it will be useful, but 
 WITHOUT ANY WARRANTY; without even the implied warranties of 
 MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
 PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along 
 with Quickshot.  If not, see <http://www.gnu.org/licenses/>.
"""
import ConfigParser
import datetime
import mimetools
import mimetypes
import os
import re
import StringIO
import urllib
import urllib2

#Steps retrieved from the server may be decorated with simple markup when
#rendered. These public constants should be used to look up the type of markup
#to apply when displaying instructions to the user.
STEP_PLAIN = 0 #: No special handling for this step.
STEP_ORDERED = 1 #: Prefix this step with a numbered bullet-point.
STEP_UNORDERED = 2 #: Prefix this step with an unordered bullet-point.
STEP_NOTE = 3 #: Mark this step as a note.
STEP_WARNING = 4 #: Mark this step as a warning.

_PROTOCOL_VERSION = '0-gamma' #: The version of the Quickshot protocol understood by this client.
_TIMEOUT_QUERY = 5 #: The number of seconds to wait before timing out a data-request.
_TIMEOUT_DOWNLOAD = 15 #: The number of seconds to wait before timing out a download-request.
_TIMEOUT_UPLOAD = 300 #: The number of seconds to wait before timing out an upload-request.

class _Session(object):
    """
    A communication-session with a Quickshot server.
    
    In essence, this just gathers a bunch of related variables and functions to
    make clients easier to script.
    """
    _address = None #: The base URL of a project.
    _language = None #: The language of screenshots to be uploaded.
    _project_languages = None #: A list of all languages the project supports.
    _interface_language = None #: The language in which the user would prefer to receive instructions.
    _public = None #: True if the project's URL has public visibility.
    _uploader_details = None #: Any user-identifying information to be sent to the server.
    
    def __init__(self, address, language, project_languages, public=False):
        """
        Builds an L{_Session} object.
        
        @type address: basestr
        @param address: The base URL of the project to access.
        @type language: basestr
        @param language: A language code, like 'en' or 'fr_CA'.
        @type project_languages: sequence
        @param project_languages: A list of all languages the project supports.
        @type public: bool
        @param public: True if the project's URL has public visibility.
        """
        self._address = address
        self._language = language
        self._project_languages = project_languages
        self._uploader_details = {
         'uploader': 'anonymous@example.org',
         'uploader-name': None,
         'uploader-homepage': None,
         'uploader-publish': 'no',
        }
        self._public = public
        
    def download_latest(self, screenshot, quality=70, thumbnail=None):
        """
        Downloads the most recent pending screenshot from the Quickshot server.
        
        @type screenshot: basestring
        @param screenshot: The name of the screenshot to retrieve.
        @type quality: int
        @param quality: The JPEG compression quality to request.
            
            Setting this to 0 will cause a PNG to be returned.
        @type thumbnail: sequence(2)
        @param thumbnail: The maximum dimensions to which the returned image
            should be scaled.
            
            Setting this to None will return the image at its normal
            resolution.
        
        @rtype: tuple(2)
        @return: The image, in binary-string form, and its content type, one of
            ('jpeg', 'png').
        
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        @raise NoScreenshotError: The Quickshot server was unable to provide a
            suitable screenshot.
        """
        url = self.get_latest_url(screenshot, quality, thumbnail)
        return self._download_screenshot(url)
        
    def download_reference(self, screenshot, quality=70, thumbnail=None):
        """
        Downloads a reference screenshot from the Quickshot server.
        
        @type screenshot: basestring
        @param screenshot: The name of the screenshot to retrieve.
        @type quality: int
        @param quality: The JPEG compression quality to request.
            
            Setting this to 0 will cause a PNG to be returned.
        @type thumbnail: sequence(2)
        @param thumbnail: The maximum dimensions to which the returned image
            should be scaled.
            
            Setting this to None will return the image at its normal
            resolution.
        
        @rtype: tuple(2)
        @return: The image, in binary-string form, and its content type, one of
            ('jpeg', 'png').
        
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        @raise NoScreenshotError: The Quickshot server was unable to provide a
            suitable screenshot.
        """
        url = self.get_reference_url(screenshot, quality, thumbnail)
        return self._download_screenshot(url)
        
    def get_address(self):
        """
        Provides the URL of the project referenced by this L{_Session}, if it
        is public.
        
        @rtype: basestring
        @return: The URL of the project referenced by this L{_Session}, or None
            if it is private.
        """
        return (self._public and self._address) or None

    def get_latest_url(self, screenshot, quality=80, thumbnail=None):
        """
        Provides the URL at which the most recent pending screenshot would
        reside.
        
        @type screenshot: basestring
        @param screenshot: The name of the screenshot to retrieve.
        @type quality: int
        @param quality: The JPEG compression quality to request.
            
            Setting this to 0 will cause a PNG to be returned.
        @type thumbnail: sequence(2)
        @param thumbnail: The maximum dimensions to which the returned image
            should be scaled.
            
            Setting this to None will return the image at its normal
            resolution.
        
        @rtype: basestr
        @return: The URL of the most recent pending screenshot.
        """
        return self._address + self._language + '/' + screenshot + '/_latest?' + self._compile_screenshot_qualifiers(quality, thumbnail)
        
    def get_reference_url(self, screenshot, quality=80, thumbnail=None):
        """
        Provides the URL at which the reference screenshot would
        reside.
        
        @type screenshot: basestring
        @param screenshot: The name of the screenshot to retrieve.
        @type quality: int
        @param quality: The JPEG compression quality to request.
            
            Setting this to 0 will cause a PNG to be returned.
        @type thumbnail: sequence(2)
        @param thumbnail: The maximum dimensions to which the returned image
            should be scaled.
            
            Setting this to None will return the image at its normal
            resolution.
        
        @rtype: basestr
        @return: The URL of the most recent pending screenshot.
        """
        return self._address + screenshot + '/_reference?' +  self._compile_screenshot_qualifiers(quality, thumbnail)
        
    def query_details(self, screenshot):
        """
        Queries a Web server for details about a specific screencap.
        
        @type screenshot: basestring
        @param screenshot: The name of the screenshot for which to retrieve
            details.
            
        @rtype: tuple(2)
        @return: The first slot of the returned tuple contains the result of
            the upload: a boolean indicating success, a short summary, and
            technical details.
            
            The second is a dict, which contains all information retrieved from
            the server.
                
                This includes C{summary:str}, C{steps:tuple}.
                
                Optional elements include C{command:tuple=None},
                C{maximize:bool=False}, C{maximize-workaround:str=None},
                C{fullscreen:bool=False}, C{sub-rect:tuple=None},
                C{sub-rect-invert-x:bool=False},
                C{sub-rect-invert-y:bool=False}.
        
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        @raise NoScreenshotError: The Quickshot server was unable to provide
            details for the requested screenshot.
        """
        meta = {}
        if self._interface_language:
            meta['interface-language'] = self._interface_language
        request = urllib2.Request(self._address + screenshot + '/_details' + ((meta and ('?' + urllib.urlencode(meta))) or ''))
        
        try:
            response = urllib2.urlopen(request, timeout=_TIMEOUT_QUERY)
            (success, summary, details) = self._parse_server_response(response.read())
            response.close()
        except Exception, e:
            raise CommunicationError(str(e))
        else:
            screenshot_details = {}
            
            if success:
                for line in details.split('\n'):
                    line = line.strip()
                    if line:
                        parts = line.split('\t', 1)
                        if parts[0] == 'summary':
                            screenshot_details['summary'] = parts[1]
                        elif parts[0] == 'steps':
                            steps = []
                            for step in parts[1].split('\t'):
                                if step.startswith('#'):
                                    steps.append((STEP_ORDERED, step[1:]))
                                elif step.startswith('*'):
                                    steps.append((STEP_UNORDERED, step[1:]))
                                elif step.startswith('='):
                                    steps.append((STEP_NOTE, step[1:]))
                                elif step.startswith('-'):
                                    steps.append((STEP_WARNING, step[1:]))
                                else:
                                    steps.append((STEP_PLAIN, step))
                            screenshot_details['steps'] = tuple(steps)
                        elif parts[0] in ('maximize', 'fullscreen', 'sub-rect-invert-x', 'sub-rect-invert-y'):
                            screenshot_details[parts[0]] = True
                        elif parts[0] == 'command':
                            screenshot_details['command'] = tuple(parts[1].split('\t'))
                        elif parts[0] == 'maximize-workaround':
                            screenshot_details['maximize-workaround'] = parts[1]
                        elif parts[0] == 'sub-rect':
                            try:
                                sub_rect = tuple([int(i) for i in parts[1].split(':', 3)])
                                if any(sub_rect) and all(sub_rect[2:]):
                                    screenshot_details['sub-rect'] = sub_rect
                            except Exception:
                                raise CommunicationError("Server provided an invalid sub-rect definition")
                                
            if not 'summary' in screenshot_details:
                raise CommunicationError("Server did not provide a screenshot summary")
            if not 'steps' in screenshot_details:
                raise CommunicationError("Server did not provide screenshot steps")
                
            return (
             (success, summary, details),
             screenshot_details,
            )
            
    def query_progress(self):
        """
        Queries a Quickshot server for information about the progress of the
        current language.
        
        @rtype: tuple(2)
        @return: The first slot of the returned tuple contains the result of
            the upload: a boolean indicating success, a short summary, and
            technical details.
            
            The second is a tuple(4), with a list of all screenshots defined on
            the server, by screenshot-name, and three lists of tuple(2)s
            containing screenshot-name and friendly-name, grouped into
            screenshots needed, pending, and accepted, for the current language.
            
                All lists are guaranteed to be ordered. To combine two lists
                and preserve order, create a C{dict} out of the lists and pull
                elements out in the sequence defined by the master list.
        
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        """
        request = urllib2.Request(self._address + self._language + '/_progress?' + urllib.urlencode({
         'interface-language': self._interface_language or self._language,
        }))
        
        try:
            response = urllib2.urlopen(request, timeout=_TIMEOUT_QUERY)
            (success, summary, details) = self._parse_server_response(response.read())
            response.close()
        except Exception, e:
            raise CommunicationError(str(e))
        else:
            master = []
            needed = []
            pending = []
            accepted = []
            
            if success:
                for line in details.split('\n'):
                    line = line.strip()
                    if line:
                        parts = line.split('\t', 2)
                        if len(parts) == 3:
                            (state, screenshot_name, friendly_name) = parts
                            master.append(screenshot_name)
                            entry = (screenshot_name, friendly_name)
                            
                            if state == 'needed':
                                needed.append(entry)
                            elif state == 'pending':
                                pending.append(entry)
                            elif state == 'accepted':
                                accepted.append(entry)
            return (
             (success, summary, details),
             (master, needed, pending, accepted),
            )
            
    def set_interface_language(self, interface_language):
        """
        Changes the interface-language associated with this L{_Session},
        altering the language in which the user will receive instructions fro,
        the server, if translations into the target language exist.
        
        If the target language is unsupported, the default language, which
        varies from project to project, will be used instead. This is the
        default behaviour if this function is not invoked.
        
        @type interface_language: basestr
        @param interface_language: The language in which the user would prefer
            to receive instructions.
        """
        self._interface_language = interface_language
        
    def set_language(self, language):
        """
        Changes the working language on the fly, provided the project supports
        the new language.
        
        @type language: basestr
        @param language: The new language code to use for server interaction.
        
        @rtype: basestr|None
        @return: The previous language code or None if the new language is
            unsupported.
        """
        if not language in self._project_languages:
            return None
            
        old_language = self._language
        self._language = language
        return old_language
        
    def set_uploader_details(
     self, uploader_email, uploader_name=None, uploader_homepage=None, uploader_publish=False
    ):
        """
        Sets the properties the user will use to mark their uploads to claim
        credit.
        
        @type uploader_email: basestr
        @param uploader_email: An e-mail address used to uniquely identify the
            user. The server is expected to take precautions to protect this
            value, to the maximum extent that makes sense.
        @type uploader_name: basestr|None
        @param uploader_name: The real name of the user.
        @type uploader_homepage: basestr|None
        @param uploader_homepage: The user's homepage URL.
        @type uploader_publish: bool
        @param uploader_publish: True if the user does not want to be anonymous.
        """
        self._uploader_details = {
         'uploader': uploader_email,
         'uploader-name': uploader_name,
         'uploader-homepage': uploader_homepage,
         'uploader-publish': (uploader_publish and 'yes') or 'no',
        }
        
    def upload_capture(self, screenshot, screenshot_name):
        """
        Uploads a screencap to the Quickshot server using HTTP POST.
        
        @type screenshot: basestring|file-like object
        @param screenshot: The path to the screenshot or an open screenshot
            file.
            
            If an open file, it must be seeked to the start of the data and
            contain nothing but screenshot data between its current position
            and the end of the file. It will be closed by this function.
            (This is a good opportunity to use anonymous tempfiles)
        @type screenshot_name: basestring
        @param screenshot_name: The name of the captured screenshot.
        
        @rtype: tuple(1)
        @return: The first slot of the returned tuple contains the result of
            the upload: a boolean indicating success, a short summary, and
            technical details.
        
        @raise IOError: The local screenshot couldn't be read.
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        """
        def _encode_content(options):
            """
            Encodes the given file and parameters into a POST body.
            
            Loosely based on code found at
            http://www.voidspace.org.uk/python/cgi.shtml#upload test_upload.py
            
            @type options: sequence
            @param options: A collection of tuples or lists, as key-value pairs.
            
            @rtype: string
            @return: The encoded payload.
            """
            boundary = '-=-' + mimetools.choose_boundary() + '-=-'
            
            lines = []
            for (key, value) in options:
                if not value is None:
                    lines += [
                     '--%(boundary)s' % {'boundary': boundary,},
                     'Content-Disposition: form-data; name="%(key)s"' % {'key': key,}, '', value,
                    ]
                    
            file_type = 'application/octet-stream'
            file_data = screenshot
            if isinstance(screenshot, basestring):
                file_type = mimetypes.guess_type(screenshot)[0] or file_type
                file_data = open(screenshot, 'rb')
                
            lines += [
             '--%(boundary)s' % {'boundary': boundary,},
             'Content-Disposition: form-data; name="screenshot"; filename="%(file)s"' % {'file': screenshot_name,},
             'Content-Type: %(file-type)s' % {'file-type': file_type,},
             'Content-Transfer-Encoding: binary',
             '', file_data.read(),
            ]
            file_data.close()
            
            lines.append('--' + boundary + '--')
            lines.append('')
            
            return ('\r\n'.join(lines), boundary)
            
        (data, boundary) = _encode_content(self._uploader_details.items())
        request = urllib2.Request(self._address + self._language + '/_upload', data, {
         'Content-Type': 'multipart/form-data; boundary=' + boundary,
         'Content-Length': str(len(data)),
        })
        
        result = None
        try:
            response = urllib2.urlopen(request, timeout=_TIMEOUT_UPLOAD)
            (success, summary, details) = self._parse_server_response(response.read())
            response.close()
        except Exception, e:
            raise CommunicationError(str(e))
        else:
            return (
             (success, summary, details),
            )

    def _compile_screenshot_qualifiers(self, quality, thumbnail):
        """
        Builds query-string components for screenshot presentation options.
        
        @type quality: int
        @param quality: The JPEG compression quality to request.
            
            Setting this to 0 will cause a PNG to be returned.
        @type thumbnail: sequence(2)
        @param thumbnail: The maximum dimensions to which the returned image
            should be scaled.
            
            Setting this to None will return the image at its normal
            resolution.
        
        @rtype: basestring
        @return: The query-parameters that reflect the given options.
        """
        qualifiers = {}
        if quality:
            qualifiers['quality'] = str(quality)
        if thumbnail:
            qualifiers['thumbnail-width'] = thumbnail[0]
            qualifiers['thumbnail-height'] = thumbnail[1]
            
        if qualifiers:
            return urllib.urlencode(qualifiers)
        return ''
        
    def _download_screenshot(self, url):
        """
        Downloads an image from the Quickshot server.
        
        @type url: basestring
        @param url: The address to be accessed.
        
        @rtype: tuple(2)
        @return: The image, in binary-string form, and its content type, one of
            C{('jpeg', 'png')}.
        
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        @raise NoScreenshotError: The Quickshot server was unable to provide a
            suitable screenshot.
        """
        request = urllib2.Request(url)
        try:
            result = urllib2.urlopen(request, timeout=_TIMEOUT_DOWNLOAD)
            data = result.read()
            
            content_type = result.info().getheader('Content-Type')
            if not content_type:
                raise CommunicationError('Content-Type not specified by server')
                
            if ';' in content_type:
                content_type = content_type[:content_type.find(';')]
                
            if content_type == 'image/jpeg':
                content_type = 'jpeg'
            elif content_type == 'image/png':
                content_type = 'png'
            else:
                raise CommunicationError('Unknown Content-Type specified by server')
                
            result.close()
            return (data, content_type)
        except urllib2.HTTPError, e:
            if e.code == 404:
                raise NoScreenshotError(e.url)
            raise CommunicationError(str(e))
        except Exception, e:
            raise CommunicationError(str(e))
            
    def _parse_server_response(self, response):
        """
        Parses the standard Quickshot server-response format.
        
        @type response: basestr
        @param response: The response received from the server.
        
        @rtype: tuple(3)
        @return: A boolean indicating whether the server's response indicated
            success, a brief summary of the result, and technical details.
            
            In the event of success, the details field will typically contain
            additional content to be parsed. Its contents should be exposed to
            users only if the result was failure.
        
        @raise CommunicationError: A problem occurred while interacting with
            the Quickshot server.
        """
        success = summary = details = None
        for line in response.split('\r'):
            parts = line.split(':', 1)
            if len(parts) == 2:
                if parts[0] == 'Result':
                    success = parts[1].strip().lower() == 'success'
                elif parts[0] == 'Summary':
                    summary = parts[1].strip()
                elif parts[0] == 'Details':
                    details = parts[1].strip()
                    
        if not success and summary and details:
            raise CommunicationError(
             "Server did not send a spec-compliant response: %(response)s" % {
              'response': response,
             }
            )
        return (success, summary, details)
        
        
def build_session_from_url(url, language):
    """
    Establishes a connection to a Quickshot server.
    
    The specified server is contacted, its protocol-version and language are
    tested for compatibility, and the most recent .qsproj file's contents are
    downloaded.
    
    The .qsproj is then parsed into a dictionary for use by client
    applications.
    
    @type url: basestr
    @param url: The base URL of the project to access.
    @type language: basestr
    @param language: A language code, like 'en' or 'fr_CA'.
    
    @rtype: tuple(2)
    @return: A _Session object used to communicate with a Quickshot server and
        a dictionary with server-specific information:
         {
         'meta': {
          'project': {
           'friendly_name': C{basestr},
           'homepage': C{basestr}, #May not appear.
          },
          'version': {
           'friendly_name': C{basestr},
           'homepage': C{basestr}, #May not appear.
          },
          'expiration-date': C{datetime.date}, #May not appear.
          'languages': C{tuple},
         },
         'config': {
          'resolution': {
           'x': int, #0 means 'any'.
           'y': int, #0 means 'any'.
          },
          'username': basestr, #May not appear.
          'display-base': basestr, #May not appear.
         },
        }
    
    @raise CommunicationError: The Quickshot server was unreachable or behaved
        erratically.
    """
    if not url.endswith('/'):
        url += '/'
        
    request = urllib2.Request(url + '/.qsproj?' + urllib.urlencode({'language': language,}))
    try:
        response = urllib2.urlopen(request, timeout=_TIMEOUT_QUERY)
        qsproj = StringIO.StringIO(response.read())
        response.close()
    except urllib2.HTTPError, e:
            if e.code == 400:
                qs_summary = None
                for line in e.readlines():
                    line = line.strip()
                    if line.startswith('QS-SUMMARY: '):
                        qs_summary = line[12:]
                        break
                if qs_summary:
                    raise CommunicationError(qs_summary)
            elif e.code == 404:
                raise CommunicationError("Unable to locate project information")
            raise CommunicationError(str(e))
    except Exception, e:
        raise CommunicationError(str(e))
    else:
        parser = ConfigParser.RawConfigParser()
        parser.readfp(qsproj)
        if not parser.sections():
            raise CommunicationError("Server provided an invalid .qsproj file")
            
        details = {
         'meta': {
          'project': {},
          'version': {},
         },
         'config': {
          'resolution': {},
         },
        }
        try:
            if not parser.get('system', 'version') == _PROTOCOL_VERSION:
                raise CommunicationError("Protocol incompatibility: client protocol-version: %(client)s; server protocol-version: %(server)s" % {
                 'client': _PROTOCOL_VERSION,
                 'server': parser.get('system', 'version'),
                })
                
            project_public = parser.get('system', 'public').lower() == 'true'
            
            details['meta']['project']['friendly-name'] = parser.get('meta', 'project-friendly-name')
            details['meta']['version']['friendly-name'] = parser.get('meta', 'version-friendly-name')
            homepage = parser.get('meta', 'project-homepage')
            if homepage:
                details['meta']['project']['homepage'] = homepage
            homepage = parser.get('meta', 'version-homepage')
            if homepage:
                details['meta']['version']['homepage'] = homepage
            expiration_date = parser.get('meta', 'expiration-date')
            if expiration_date:
                details['meta']['expiration-date'] = datetime.date(*[int(i) for i in expiration_date.split('-')])
            details['meta']['languages'] = tuple(parser.get('meta', 'languages').split(','))
            
            details['config']['resolution']['x'] = max(0, int(parser.get('config', 'resolution-x')))
            details['config']['resolution']['y'] = max(0, int(parser.get('config', 'resolution-y')))
            username = parser.get('config', 'username')
            if username:
                details['config']['username'] = username
            display_name = parser.get('config', 'display-name')
            if display_name:
                details['config']['display-name'] = display_name
        except CommunicationError, e:
            raise e
        except Exception, e:
            raise CommunicationError("Server provided an incomplete .qsproj file: %(error)s" % {
             'error': str(e),
            })
        else:
            session = _Session(url, language, details['meta']['languages'], public=project_public)
            return (session, details)
            
def get_url_from_qsproj(path):
    """
    Extracts the URL of the server from the .qsproj file, for use by
    L{build_session_from_url} to pull the most recent data.
    
    @type path: basestr
    @param path: The location of a .qsproj file in the user's filesystem.
    
    @rtype: basestr
    @return: The project-URL of the Quickshot server.
    
    @raise IOError: The .qsproj file could not be processed.
    """
    parser = ConfigParser.RawConfigParser()
    if not parser.read(path):
        raise IOError("Unable to parse .qsproj file at '%(path)s'" % {
         'path': path,
        })
        
    if not parser.has_option('system', 'url'):
        raise IOError("Invalid or corrupt .qsproj file at '%(path)s'" % {
         'path': path,
        })
        
    return parser.get('system', 'url')
    
class Error(Exception):
    """
    The base class from which all exceptions native to this module inherit.
    """
    def __init__(self, message):
        """
        Creates a generic Error instance.
        
        @type message: basestring
        @param message: The message to be relayed.
        """
        Exception.__init__(self, message)
        
class CommunicationError(Error):
    """
    Signifies that the Quickshot server is unavailable or misbehaving.
    """
    pass
    
class NoScreenshotError(Error):
    """
    Signifies that the Quickshot server could not provide details for the
    specified screenshot.
    """
    pass
    
