import datetime
import logging
import re

from pylons import config, url, request, response, session, tmpl_context as c
from pylons.controllers.util import abort, redirect_to

from sqlalchemy.exc import IntegrityError
import sqlalchemy.sql

from quickshot.lib.base import BaseController, render
import quickshot.lib.helpers as h
import quickshot.model as model
import quickshot.model.meta as meta

log = logging.getLogger(__name__)

class ManageProjectController(BaseController):
    def index(self, project_family, project):
        if not self._load_information(project_family, project):
            h.unauthorized()
            
        return render('/manage_project.mako')
        
    def language_add(self, project_family, project):
        if not self._load_information(project_family, project):
            h.unauthorized()
            
        language_code = request.params.get('language', '').strip()
        if not self._check_language(language_code):
            redirect_to(c.current_page)
            
        #Create the new entry
        new_project_language = model.ProjectLanguage()
        new_project_language.p_short_name = project_family
        new_project_language.p_version = project
        new_project_language.language_code = language_code
        
        meta.Session.add(new_project_language)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given language already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to add language", details),
            )
        else:
            session['success_report'] = (
             ("Added version language", (
              "Language: %(language)s" % {
               'language': language_code,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def language_remove(self, project_family, project):
        if not self._load_information(project_family, project):
            h.unauthorized()
            
        language_code = request.params.get('language', '').strip()
        if not self._check_language(language_code):
            redirect_to(c.current_page)
            
        #Delete the entry
        try:
            meta.Session.query(
             model.ProjectLanguage
            ).filter(
             model.ProjectLanguage.p_short_name == project_family
            ).filter(
             model.ProjectLanguage.p_version == project
            ).filter(
             model.ProjectLanguage.language_code == language_code
            ).delete()
            
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to remove language", (
              h.screen_error(e),
             )),
            )
        else:
            session['success_report'] = (
             ("Removed language", (
              "Language: %(language)s" % {
               'language': language_code,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def screenshot_create(self, project_family, project):
        if not self._load_information(project_family, project):
            h.unauthorized()
            
        friendly_name = request.params.get('friendly-name', '').strip()
        screenshot_name = request.params.get('screenshot-name', '').strip()
        if not friendly_name or not screenshot_name:
            details = []
            if not friendly_name:
                details.append("Friendly name not supplied")
            if not screenshot_name:
                details.append("Screenshot name not supplied")
                
            session['problem_report'] = (
             ("Insufficient information provided", details),
            )
            session.save()
            redirect_to(c.current_page)
            
        if not self._check_screenshot_name(screenshot_name):
            redirect_to(c.current_page)
            
        #Create the new entry
        new_project_screenshot = model.ProjectScreenshot()
        new_project_screenshot.p_short_name = unicode(project_family)
        new_project_screenshot.p_version = unicode(project)
        new_project_screenshot.screenshot_name = screenshot_name
        new_project_screenshot.friendly_name = friendly_name
        new_project_screenshot.steps = u"No steps defined"
        new_project_screenshot.display_order = (meta.Session.query(
         model.ProjectScreenshot.display_order
        ).filter(
         model.ProjectScreenshot.p_short_name == project_family
        ).filter(
         model.ProjectScreenshot.p_version == project
        ).filter(
         model.ProjectScreenshot.display_order == sqlalchemy.sql.func.max(model.ProjectScreenshot.display_order).select()
        ).scalar() or 0) + 1
        
        meta.Session.add(new_project_screenshot)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given screenshot already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to create new screenshot", details),
            )
        else:
            session['success_report'] = (
             ("Created screenshot", (
              "Friendly name: %(friendly-name)s" % {
               'friendly-name': friendly_name,
              },
              "Screenshot name: %(screenshot-name)s" % {
               'screenshot-name': screenshot_name,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def screenshot_modify(self, project_family, project):
        if not self._load_information(project_family, project):
            h.unauthorized()
            
        friendly_name = request.params.get('friendly-name', '').strip()
        screenshot_name = request.params.get('screenshot-name', '').strip()
        if not friendly_name or not screenshot_name:
            details = []
            if not friendly_name:
                details.append("Friendly name not supplied")
            if not screenshot_name:
                details.append("Screenshot name not supplied")
                
            session['problem_report'] = (
             ("Insufficient information provided", details),
            )
            session.save()
            redirect_to(c.current_page)
            
        if not self._check_screenshot_name(screenshot_name):
            redirect_to(c.current_page)
            
        screenshot = meta.Session.query(
         model.ProjectScreenshot
        ).filter(
         model.ProjectScreenshot.p_short_name == project_family
        ).filter(
         model.ProjectScreenshot.p_version == project
        ).filter(
         model.ProjectScreenshot.screenshot_name == screenshot_name
        ).first()
        if not screenshot:
            session['problem_report'] = (
             ("Specified screenshot was not found", ()),
            )
            session.save()
            redirect_to(c.current_page)
            
        problems = []
        #Update the display order
        try:
            display_slot = request.params.get('display-order', '')
            if display_slot and not int(display_slot) == screenshot.display_order and int(display_slot) >= 0:
                display_slot = int(display_slot)
                if display_slot == 0: #0 is a special value, meaning end-of-list.
                    display_slot = 1000000000 #If a project has one billion screenshots, well...
                    
                other_screenshots = meta.Session.query(
                 model.ProjectScreenshot
                ).filter(
                 model.ProjectScreenshot.p_short_name == project_family
                ).filter(
                 model.ProjectScreenshot.p_version == project
                ).filter(
                 model.ProjectScreenshot.screenshot_name != screenshot_name
                ).order_by(
                 model.ProjectScreenshot.display_order.asc()
                ).with_lockmode('update').all()
                
                if display_slot < screenshot.display_order: #Moving up.
                    compacting_display_slot = display_slot
                    for other_screenshot in other_screenshots:
                        if other_screenshot.display_order >= display_slot:
                            compacting_display_slot += 1
                            if not other_screenshot.display_order == compacting_display_slot:
                                other_screenshot.display_order = compacting_display_slot
                    screenshot.display_order = display_slot
                else: #Moving down.
                    initial_display_slot = compacting_display_slot = screenshot.display_order
                    screenshot_target = 0
                    for other_screenshot in other_screenshots:
                        if other_screenshot.display_order >= initial_display_slot:
                            if not screenshot_target and other_screenshot.display_order == display_slot:
                                screenshot_target = compacting_display_slot
                                compacting_display_slot += 1
                            if not other_screenshot.display_order == compacting_display_slot:
                                other_screenshot.display_order = compacting_display_slot
                            compacting_display_slot += 1
                    screenshot.display_order = screenshot_target or compacting_display_slot
        except ValueError:
            problems.append(("Unable to parse list position", ()))
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to update screenshot display order", (h.screen_error(e),)),
            )
            session.save()
            redirect_to(c.current_page)
            
        #Update the entry
        screenshot.friendly_name = unicode(friendly_name)
        screenshot.steps = unicode('\n'.join([line.strip() for line in request.params.get('steps', 'No steps defined').splitlines()]))
        screenshot.command = unicode('\n'.join([line for line in [l.strip() for l in request.params.get('command', '').splitlines()] if line])) or None
        screenshot.maximize = bool(request.params.get('maximize', False))
        screenshot.maximize_workaround = unicode(request.params.get('maximize-workaround', '').strip()) or None
        screenshot.fullscreen = bool(request.params.get('fullscreen', False))
        
        (screenshot.sub_rect_left, success) = self._parse_resolution('sub-rect-left')
        if not success:
            problems.append(("Unable to parse left offset", ()))
        (screenshot.sub_rect_top, success) = self._parse_resolution('sub-rect-top')
        if not success:
            problems.append(("Unable to parse top offset", ()))
        (screenshot.sub_rect_width, success) = self._parse_resolution('sub-rect-width')
        if not success:
            problems.append(("Unable to parse width", ()))
        (screenshot.sub_rect_height, success) = self._parse_resolution('sub-rect-height')
        if not success:
            problems.append(("Unable to parse height", ()))
            
        screenshot.sub_rect_invert_x = bool(request.params.get('sub-rect-invert-x', False))
        screenshot.sub_rect_invert_y = bool(request.params.get('sub-rect-invert-y', False))
        screenshot.active = bool(request.params.get('active', False))
        
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            problems.append(("Unable to update screenshot", (h.screen_error(e),)))
        else:
            session['success_report'] = (
             ("Updated screenshot", (
              "Friendly name: %(friendly-name)s" % {
               'friendly-name': friendly_name,
              },
              "Active: %(active)s" % {
               'active': screenshot.active,
              },
              "Steps: %(steps)s" % {
               'steps': repr(screenshot.steps),
              },
              "Command: %(command)s" % {
               'command': repr(screenshot.command),
              },
              "Maximize: %(maximize)s" % {
               'maximize': screenshot.maximize,
              },
              "Maximize workaround process name: %(maximize-workaround)s" % {
               'maximize-workaround': screenshot.maximize_workaround,
              },
              "Capture full desktop: %(fullscreen)s" % {
               'fullscreen': screenshot.fullscreen,
              },
              "Cropping rectangle: %(sub-rect)s" % {
               'sub-rect': (screenshot.sub_rect_left, screenshot.sub_rect_top, screenshot.sub_rect_width, screenshot.sub_rect_height),
              },
              "Crop relative to right edge: %(invert-x)s" % {
               'invert-x': screenshot.sub_rect_invert_x,
              },
              "Crop relative to bottom edge: %(invert-y)s" % {
               'invert-y': screenshot.sub_rect_invert_y,
              },
              "Active: %(active)s" % {
               'active': screenshot.active,
              },
             )),
            )
        finally:
            if problems:
                session['problem_report'] = problems
            session.save()
        redirect_to(c.current_page)
        
    def settings(self, project_family, project):
        if not self._load_information(project_family, project):
            h.unauthorized()
            
        friendly_name = request.params.get('friendly-name', '').strip()
        if not friendly_name:
            session['problem_report'] = (
             ("Insufficient information provided", (
              "Friendly name not supplied",
             )),
            )
            session.save()
            redirect_to(c.current_page)
            
        username = request.params.get('username', '').strip()
        if not username:
            username = None
        elif not re.match(r'^[a-z_][a-z0-9_-]{0,254}$', username):
            session['problem_report'] = (
             ("Invalid username provided", (
              "Usernames must be all-lower-case, start with an alpha-character or underscore, and contain only alphanumerics, underscores, and hyphens",
             )),
            )
            session.save()
            redirect_to(c.current_page)
        else:
            username = unicode(username)
            
        problems = []
        #Update the entry
        c.project.friendly_name = unicode(friendly_name)
        c.project.username = username
        c.project.display_name = unicode(request.params.get('display-name', '').strip()) or None
        c.project.public = not bool(request.params.get('private', False))
        c.project.active = bool(request.params.get('active', False))
        c.project.homepage = unicode(request.params.get('homepage', '').strip()) or None
        
        (c.project.resolution_x, success) = self._parse_resolution('resolution-x', default=0)
        if not success:
            problems.append(("Unable to parse X-resolution", ()))
        (c.project.resolution_y, success) = self._parse_resolution('resolution-y', default=0)
        if not success:
            problems.append(("Unable to parse Y-resolution", ()))
            
        expiration_date = request.params.get('expiration-date', '').strip()
        c.project.expiration_date = None
        try:
            if expiration_date:
                c.project.expiration_date = datetime.date(*[int(i) for i in expiration_date.split('-')])
        except Exception, e:
            problems.append(("Unable to interpret date", ("Values must be in yyyy-mm-dd format", str(e))))
            
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            problems.append(("Unable to update version settings", (h.screen_error(e),)))
        else:
            session['success_report'] = (
             ("Updated project settings for %(version)s" % {
              'version': c.project.version,
             }, (
              "Friendly name: %(friendly-name)s" % {
               'friendly-name': c.project.friendly_name,
              },
              "Private: %(private)s" % {
               'private': not c.project.public,
              },
              "Active: %(active)s" % {
               'active': c.project.active,
              },
              "Deadline: %(deadline)s" % {
               'deadline': c.project.expiration_date,
              },
              "X-resolution: %(resolution-x)s" % {
               'resolution-x': c.project.resolution_x or 'any',
              },
              "Y-resolution: %(resolution-y)s" % {
               'resolution-y': c.project.resolution_y or 'any',
              },
              "Username: %(username)s" % {
               'username': c.project.username,
              },
              "Display: %(display-name)s" % {
               'display-name': c.project.display_name,
              },
              "Homepage: %(homepage)s" % {
               'homepage': c.project.homepage,
              },
             )),
            )
        finally:
            if problems:
                session['problem_report'] = problems
            session.save()
        redirect_to(c.current_page)
        
    def _check_language(self, language):
        if not h.check_presence(language, "Language"):
            return False
            
        if not re.match(config['path.valid_language'], language):
            session['problem_report'] = (
             ("Invalid language", (
              "Languages must begin with two lower-case letters and optionally specify a dialect (%(regexp)s)" % {
               'regexp': config['path.valid_language'],
              },
              "Valid examples: 'en', 'en_GB'",
             )),
            )
            session.save()
            return False
        return True
        
    def _check_screenshot_name(self, screenshot_name):
        if not re.match(config['path.valid_name'], screenshot_name):
            session['problem_report'] = (
             ("Invalid screenshot name", (
              "Screenshot names must be all-lower-case, start with an alphanumeric, and use only '-' as a separator (%(regexp)s)" % {
               'regexp': config['path.valid_name'],
              },
             )),
            )
            session.save()
            return False
        return True
        
    def _load_information(self, project_family, project):
        c.current_page = url(str(''.join(('/', project_family, '/', project, '/manage_settings'))))
        c.admin = h.assert_owner(project_family)
        c.project = h.get_project_details(project_family, project)
        
        return c.project and c.admin
        
    def _parse_resolution(self, resolution_name, default=None):
        resolution = request.params.get(resolution_name, '')
        if not resolution:
            if default is None:
                return(default, True)
            return (default, False)
            
        try:
            return (min(32767, max(0, int(resolution))), True)
        except Exception:
            return (default, False)
            
