import logging
import re

from pylons import config, url, request, response, session, tmpl_context as c
from pylons.controllers.util import abort, redirect_to

from sqlalchemy.exc import IntegrityError

from quickshot.lib.base import BaseController, render
import quickshot.lib.helpers as h
import quickshot.model as model
import quickshot.model.meta as meta

log = logging.getLogger(__name__)

class ManageProjectFamilyController(BaseController):
    def index(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        c.projects = meta.Session.query(
         model.Project
        ).filter(
         model.Project.pf_short_name == project_family
        ).order_by(
         model.Project.active.desc(),
         model.Project.friendly_name.asc(),
         model.Project.version.asc()
        ).all()
        
        return render('/manage_project_family.mako')
        
    def create(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        friendly_name = request.params.get('friendly-name', '').strip()
        version = request.params.get('version', '').strip()
        if not friendly_name or not version:
            details = []
            if not friendly_name:
                details.append("Friendly name not supplied")
            if not version:
                details.append("Version not supplied")
                
            session['problem_report'] = (
             ("Insufficient information provided", details),
            )
            session.save()
            redirect_to(c.current_page)
            
        if not self._check_version(version):
            redirect_to(c.current_page)
            
        #Create the new entry
        new_project = model.Project()
        new_project.pf_short_name = unicode(project_family)
        new_project.version = version
        new_project.friendly_name = friendly_name
        new_project.resolution_x = 0
        new_project.resolution_y = 0
        
        meta.Session.add(new_project)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given version already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to create new version", details),
            )
        else:
            session['success_report'] = (
             ("Created version", (
              "Friendly name: %(friendly-name)s" % {
               'friendly-name': friendly_name,
              },
              "Version: %(version)s" % {
               'version': version,
              },
              "TODO: Create directory structure",
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def language_add(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        language_code = request.params.get('language', '').strip()
        if not self._check_language(language_code):
            redirect_to(c.current_page)
            
        #Create the new entry
        new_project_family_language = model.ProjectFamilyLanguage()
        new_project_family_language.pf_short_name = project_family
        new_project_family_language.language_code = language_code
        
        meta.Session.add(new_project_family_language)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given language already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to add language", details),
            )
        else:
            session['success_report'] = (
             ("Added project language", (
              "Language: %(language)s" % {
               'language': language_code,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def language_remove(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        language_code = request.params.get('language', '').strip()
        if not self._check_language(language_code):
            redirect_to(c.current_page)
            
        #Delete the entry
        try:
            meta.Session.query(
             model.ProjectFamilyLanguage
            ).filter(
             model.ProjectFamilyLanguage.pf_short_name == project_family
            ).filter(
             model.ProjectFamilyLanguage.language_code == language_code
            ).delete()
            
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to remove language", (
              h.screen_error(e),
             )),
            )
        else:
            session['success_report'] = (
             ("Removed language", (
              "Language: %(language)s" % {
               'language': language_code,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def owner_grant(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        openid = request.params.get('openid', '').strip()
        if not self._check_openid(openid):
            redirect_to(c.current_page)
            
        try:
            openid = h.normalise_openid(openid)
        except ValueError, e:
            session['problem_report'] = (
             ("The OpenID you provided was in a format that could not be understood", (
              str(e),
             )),
            )
            session.save()
            redirect_to(c.current_url)
            
        #Create the new entry
        new_project_family_owner = model.ProjectFamilyOwner()
        new_project_family_owner.pf_short_name = project_family
        new_project_family_owner.openid = openid
        
        meta.Session.add(new_project_family_owner)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given owner already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to add owner", details),
            )
        else:
            session['success_report'] = (
             ("Added project owner", (
              "Owner: %(openid)s" % {
               'openid': openid,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def owner_revoke(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        openid = request.params.get('openid', '').strip()
        if not self._check_openid(openid):
            redirect_to(c.current_page)
            
        #Delete the entry
        try:
            meta.Session.query(
             model.ProjectFamilyOwner
            ).filter(
             model.ProjectFamilyOwner.pf_short_name == project_family
            ).filter(
             model.ProjectFamilyOwner.openid == openid
            ).delete()
            
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to remove owner", (
              h.screen_error(e),
             )),
            )
        else:
            session['success_report'] = (
             ("Removed owner", (
              "Owner: %(openid)s" % {
               'openid': openid,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def settings(self, project_family):
        if not self._load_information(project_family):
            h.unauthorized()
            
        friendly_name = request.params.get('friendly-name', '').strip()
        if not friendly_name:
            session['problem_report'] = (
             ("Insufficient information provided", (
              "Friendly name not supplied",
             )),
            )
            session.save()
            redirect_to(c.current_page)
            
        #Update the entry
        c.project_family.friendly_name = unicode(friendly_name)
        c.project_family.public = not bool(request.params.get('private', False))
        c.project_family.active = bool(request.params.get('active', False))
        c.project_family.homepage = unicode(request.params.get('homepage', '').strip()) or None
        
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to update project settings", (
              h.screen_error(e),
             )),
            )
        else:
            session['success_report'] = (
             ("Updated project settings for %(short-name)s" % {
              'short-name': c.project_family.short_name,
             }, (
              "Friendly name: %(friendly-name)s" % {
               'friendly-name': c.project_family.friendly_name,
              },
              "Private: %(private)s" % {
               'private': not c.project_family.public,
              },
              "Active: %(active)s" % {
               'active': c.project_family.active,
              },
              "Homepage: %(homepage)s" % {
               'homepage': c.project_family.homepage,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def _check_language(self, language):
        if not h.check_presence(language, "Language"):
            return False
            
        if not re.match(config['path.valid_language'], language):
            session['problem_report'] = (
             ("Invalid language", (
              "Languages must begin with two lower-case letters and optionally specify a dialect (%(regexp)s)" % {
               'regexp': config['path.valid_language'],
              },
              "Valid examples: 'en', 'en_GB'",
             )),
            )
            session.save()
            return False
        return True
        
    def _check_openid(self, openid):
        return h.check_presence(openid, "OpenID")
        
    def _check_version(self, version):
        if not re.match(config['path.valid_name'], version):
            session['problem_report'] = (
             ("Invalid version", (
              "Versions must be all-lower-case, start with an alphanumeric, and use only '-' as a separator (%(regexp)s)" % {
               'regexp': config['path.valid_name'],
              },
             )),
            )
            session.save()
            return False
        return True
        
    def _load_information(self, project_family):
        c.current_page = url(str(''.join(('/', project_family, '/manage_settings'))))
        c.admin = h.assert_owner(project_family)
        c.project_family = h.get_project_family_details(project_family)
        
        return c.project_family and c.admin
        
