import logging
import os
import re
import time

from pylons import config, url, request, response, session, tmpl_context as c
from pylons.controllers.util import abort, redirect_to

from sqlalchemy.exc import IntegrityError
import sqlalchemy.sql

from quickshot.lib.base import BaseController, render
import quickshot.lib.helpers as h
import quickshot.model as model
import quickshot.model.meta as meta

log = logging.getLogger(__name__)

class ManageProjectScreenshotApprovalController(BaseController):
    def index(self, project_family, project, language, screenshot):
        if not self._load_information(project_family, project, language, screenshot):
            h.unauthorized()
            
        def _extract_timestamps(
         filenames,
         regexp=re.compile(r"^.+?@(?P<timestamp>\d+)\." + config['quickshot.screenshot_type'] + "$")
        ):
            timestamps = []
            for filename in filenames:
                match = regexp.match(filename)
                if match:
                    timestamps.append((match.group('timestamp'), filename))
            return timestamps
            
        pending_screenshots = _extract_timestamps(h.order_screenshots(
         c.project, language, config['quickshot.data_state_pending'], screenshot
        ))
        rejected_screenshots = _extract_timestamps(h.order_screenshots(
         c.project, language, config['quickshot.data_state_rejected'], screenshot
        ))
        
        c.pending_screenshots = []
        for (timestamp, filename) in pending_screenshots:
            uploader_details = self._get_uploader_query(filename).first()
            c.pending_screenshots.append((timestamp, uploader_details))
        c.rejected_screenshots = [timestamp for (timestamp, filename) in rejected_screenshots]
        
        c.has_accepted_screenshot = bool(h.find_accepted_screenshot(c.project, language, screenshot))
        c.has_reference_screenshot = bool(h.find_reference_screenshot(c.project, None, screenshot))
        
        return render('/manage_project_screenshot_approval.mako')
        
    def accept(self, project_family, project, language, screenshot):
        if not self._load_information(project_family, project, language, screenshot):
            h.unauthorized()
            
        base_path = h.get_screenshots_base_path(c.project, language=language)
        screenshot_id = request.params.get('screenshot-id', '').split(':', 1)
        if len(screenshot_id) == 2:
            old_filename = screenshot + '@' + screenshot_id[1] + '.' + config['quickshot.screenshot_type']
            new_filename = screenshot + '.' + config['quickshot.screenshot_type']
            if screenshot_id[0] == 'pending':
                self._reject_accepted(base_path)
                
                source_path = os.path.join(base_path, config['quickshot.data_state_pending'], old_filename)
                destination_path = os.path.join(base_path, config['quickshot.data_state_accepted'], new_filename)
                
                self._move_file(source_path, destination_path)
                self._process_upload_credit(old_filename)
                self._reject_all_pending(base_path)
            elif screenshot_id[0] == 'rejected':
                self._reject_accepted(base_path)
                
                source_path = os.path.join(base_path, config['quickshot.data_state_rejected'], old_filename)
                destination_path = os.path.join(base_path, config['quickshot.data_state_accepted'], new_filename)
                
                self._move_file(source_path, destination_path)
                self._reject_all_pending(base_path)
            else:
                h.unsupported()
        else:
            h.unsupported()
            
        session['success_report'] = (
         ("Accepted screenshot", (
          "Language: %(language)s" % {
           'language': language,
          },
          "Screenshot: %(screenshot)s" % {
           'screenshot': screenshot,
          },
         )),
        )
        session.save()
        
        redirect_to(url(str(''.join(('/', project_family, '/', project, '/', language)))))
        
    def reject(self, project_family, project, language, screenshot):
        if not self._load_information(project_family, project, language, screenshot):
            h.unauthorized()
            
        base_path = h.get_screenshots_base_path(c.project, language=language)
        screenshot_id = request.params.get('screenshot-id', '').split(':', 1)
        if screenshot_id[0] == 'accepted':
            self._reject_accepted(base_path)
        elif screenshot_id[0] == 'pending' and len(screenshot_id) == 2:
            filename = screenshot + '@' + screenshot_id[1] + '.' + config['quickshot.screenshot_type']
            source_path = os.path.join(base_path, config['quickshot.data_state_pending'], filename)
            destination_path = os.path.join(base_path, config['quickshot.data_state_rejected'], filename)
            
            self._move_file(source_path, destination_path)
            self._process_upload_credit(filename, credit=False)
        else:
            h.unsupported()
            
        pending_screenshots = h.order_screenshots(c.project, language, config['quickshot.data_state_pending'], screenshot)
        details = [
         "Language: %(language)s" % {
          'language': language,
         },
         "Screenshot: %(screenshot)s" % {
          'screenshot': screenshot,
         },
        ]
        if not pending_screenshots:
            details.append("No pending candidates remain")
        session['success_report'] = (
         ("Rejected screenshot", details),
        )
        session.save()
        
        if not pending_screenshots:
            redirect_to(c.current_page)
        else:
            redirect_to(url(str(''.join(('/', project_family, '/', project, '/', language)))))
            
    def _get_uploader_query(self, filename):
        return meta.Session.query(
         model.ProjectScreenshotUpload
        ).filter(
         model.ProjectScreenshotUpload.p_short_name == c.project.pf_short_name
        ).filter(
         model.ProjectScreenshotUpload.p_version == c.project.version
        ).filter(
         model.ProjectScreenshotUpload.language_code == c.language
        ).filter(
         model.ProjectScreenshotUpload.filename == filename
        )
        
    def _load_information(self, project_family, project, language, screenshot):
        c.current_page = url(str(''.join(('/', project_family, '/', project, '/', language, '/', screenshot))))
        c.admin = h.assert_owner(project_family)
        c.project = h.get_project_details(project_family, project)
        c.language = language
        c.screenshot = h.get_screenshot(c.project, screenshot)
        if not c.screenshot:
            h.unsupported()
            
        return c.project and c.admin
        
    def _move_file(self, source, destination):
        base_path = os.path.dirname(destination)
        if not os.path.isdir(base_path):
            old_umask = os.umask(0)
            try:
                os.makedirs(base_path, 0700)
            except Exception, e:
                session['problem_report'] = (
                 ("Unable to create destination path", (
                  h.screen_error(e),
                 )),
                )
                session.save()
                redirect_to(c.current_page)
            finally:
                os.umask(old_umask)
                
        try:
            os.rename(source, destination)
        except Exception, e:
            session['problem_report'] = (
             ("Unable to move file", (
              h.screen_error(e),
             )),
            )
            session.save()
            redirect_to(c.current_page)
            
    def _process_upload_credit(self, filename, credit=True):
        uploader_query = self._get_uploader_query(filename)
        
        if credit:
            uploader = uploader_query.first()
            if uploader:
                uploader_record =  meta.Session.query(
                 model.ProjectScreenshotApproval
                ).filter(
                 model.ProjectScreenshotApproval.p_short_name == c.project.pf_short_name
                ).filter(
                 model.ProjectScreenshotApproval.p_version == c.project.version
                ).filter(
                 model.ProjectScreenshotApproval.uploader == uploader.uploader
                ).first()
                
                new_entry = not uploader_record
                if new_entry:
                    uploader_record =  model.ProjectScreenshotApproval()
                    uploader_record.p_short_name = c.project.pf_short_name
                    uploader_record.p_version = c.project.version
                    uploader_record.uploader = uploader.uploader
                    
                uploader_record.uploader_name = uploader.uploader_name
                uploader_record.uploader_homepage = uploader.uploader_homepage
                uploader_record.uploader_publish = uploader.uploader_publish
                uploader_record.submission_count = (uploader_record.submission_count and uploader_record.submission_count + 1) or 1
                
                if new_entry:
                    meta.Session.add(uploader_record)
                    
                try:
                    meta.Session.commit()
                except Exception, e:
                    meta.Session.rollback()
                    
                    details = []
                    if type(e) is IntegrityError:
                        details.append("The uploader's record already exists")
                    else:
                        details.append(h.screen_error(e))
                        
                    session['problem_report'] = (
                     ("Unable to update uploader-record", details),
                    )
                    session.save()
                    
        try:
            uploader_query.delete()
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to remove credit record", (
              "Filename: %(filename)s" % {
               'filename': filename,
              },
              h.screen_error(e),
             )),
            )
            session.save()
            
    def _reject_accepted(self, base_path):
        source_path = os.path.join(
         base_path,
         config['quickshot.data_state_accepted'],
         c.screenshot.screenshot_name + '.' + config['quickshot.screenshot_type']
        )
        destination_path = os.path.join(
         base_path,
         config['quickshot.data_state_rejected'],
         c.screenshot.screenshot_name + '@' + str(int(time.time())) + '.' + config['quickshot.screenshot_type']
        )
        
        if os.path.isfile(source_path):
            self._move_file(source_path, destination_path)
            
    def _reject_all_pending(self, base_path):
        for filename in h.order_screenshots(c.project, c.language, config['quickshot.data_state_pending'], c.screenshot.screenshot_name):
            source_path = os.path.join(base_path, config['quickshot.data_state_pending'], filename)
            destination_path = os.path.join(base_path, config['quickshot.data_state_rejected'], filename)
            
            if os.path.isfile(source_path):
                self._move_file(source_path, destination_path)
            self._process_upload_credit(filename, credit=False)
            
