import logging
import re

from pylons import config, url, request, response, session, tmpl_context as c
from pylons.controllers.util import abort, redirect_to

from sqlalchemy.exc import IntegrityError

from quickshot.lib.base import BaseController, render
import quickshot.lib.helpers as h
import quickshot.model as model
import quickshot.model.meta as meta

log = logging.getLogger(__name__)

class ManageServerController(BaseController):
    def index(self):
        if not self._check_authorization():
            h.unauthorized()
            
        c.project_families = meta.Session.query(
         model.ProjectFamily
        ).all()
        return render('/manage_server.mako')
        
    def create(self):
        if not self._check_authorization():
            h.unauthorized()
            
        friendly_name = request.params.get('friendly-name', '').strip()
        short_name = request.params.get('short-name', '').strip()
        if not friendly_name or not short_name:
            details = []
            if not friendly_name:
                details.append("Friendly name not supplied")
            if not short_name:
                details.append("Short name not supplied")
                
            session['problem_report'] = (
             ("Insufficient information provided", details),
            )
            session.save()
            redirect_to(c.current_page)
            
        if not self._check_short_name(short_name):
            redirect_to(c.current_page)
            
        #Create the new entry
        new_project_family = model.ProjectFamily()
        new_project_family.short_name = short_name
        new_project_family.friendly_name = friendly_name
        
        meta.Session.add(new_project_family)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given short name already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to create new project", details),
            )
        else:
            session['success_report'] = (
             ("Created project", (
              "Friendly name: %(friendly-name)s" % {
               'friendly-name': friendly_name,
              },
              "Short name: %(short-name)s" % {
               'short-name': short_name,
              },
              "Before it can be properly used, you must assign at least one project owner",
              "TODO: Create directory structure",
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def grant(self):
        if not self._check_authorization():
            h.unauthorized()
            
        short_name = request.params.get('short-name', '').strip()
        openid = request.params.get('openid', '').strip()
        if not self._check_openid_targeting(short_name, openid):
            redirect_to(c.current_page)
            
        try:
            openid = h.normalise_openid(openid)
        except ValueError, e:
            session['problem_report'] = (
             ("The OpenID you provided was in a format that could not be understood", (
              str(e),
             )),
            )
            session.save()
            redirect_to(c.current_page)
            
        if not self._check_short_name(short_name):
            redirect_to(c.current_page)
            
        #Create the new entry
        new_project_family_owner = model.ProjectFamilyOwner()
        new_project_family_owner.pf_short_name = short_name
        new_project_family_owner.openid = openid
        
        meta.Session.add(new_project_family_owner)
        try:
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            details = []
            if type(e) is IntegrityError:
                details.append("The given owner already exists")
            else:
                details.append(h.screen_error(e))
                
            session['problem_report'] = (
             ("Unable to add owner", details),
            )
        else:
            session['success_report'] = (
             ("Added project owner", (
              "Project short name: %(short-name)s" % {
               'short-name': short_name,
              },
              "Owner: %(openid)s" % {
               'openid': openid,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def revoke(self):
        if not self._check_authorization():
            h.unauthorized()
            
        short_name = request.params.get('short-name', '').strip()
        openid = request.params.get('openid', '').strip()
        if not self._check_openid_targeting(short_name, openid):
            redirect_to(c.current_page)
            
        if not self._check_short_name(short_name):
            redirect_to(c.current_page)
            
        #Delete the entry
        try:
            meta.Session.query(
             model.ProjectFamilyOwner
            ).filter(
             model.ProjectFamilyOwner.pf_short_name == short_name
            ).filter(
             model.ProjectFamilyOwner.openid == openid
            ).delete()
            
            meta.Session.commit()
        except Exception, e:
            meta.Session.rollback()
            
            session['problem_report'] = (
             ("Unable to remove owner", (
              h.screen_error(e),
             )),
            )
        else:
            session['success_report'] = (
             ("Removed owner", (
              "Project short name: %(short-name)s" % {
               'short-name': short_name,
              },
              "Owner: %(openid)s" % {
               'openid': openid,
              },
             )),
            )
        finally:
            session.save()
        redirect_to(c.current_page)
        
    def _check_authorization(self):
        c.current_page = url('/manage_admin')
        
        return session.get('openid', None) == config['admin.openid']
        
    def _check_openid_targeting(self, short_name, openid):
        return h.check_presence(short_name, "Short name") and h.check_presence(openid, "OpenID")
        
    def _check_short_name(self, short_name):
        if not re.match(config['path.valid_name'], short_name):
            session['problem_report'] = (
             ("Invalid short name", (
              "Short names must be all-lower-case, start with an alphanumeric, and use only '-' as a separator (%(regexp)s)" % {
               'regexp': config['path.valid_name'],
              },
             )),
            )
            session.save()
            return False
        return True
        
