import imghdr
import logging
import os
import re
import shutil
import time

from pylons import config, request, response
from pylons.controllers.util import abort

from quickshot.lib.base import BaseController
import quickshot.lib.helpers as h
import quickshot.model as model
import quickshot.model.meta as meta

log = logging.getLogger(__name__)

class ScreenshotUploadController(BaseController):
    def index(self, project_family, project, language):
        project_details = h.get_project_details(project_family, project)
        
        if not project_details or not project_details.is_active:
            abort(404, "Project not found")
        response.content_type = 'text/plain'
        
        #Extract uploader details.
        uploader = unicode(request.params.get('uploader', ""))
        if not re.match("^\S+@\S+\.\S+$", uploader): #Not a strict check.
            return h.serve_error(
             "Invalid uploader",
             "The value provided does not look like an e-mail address"
            )
        uploader_name = unicode(request.params.get('uploader-name', '')) or None
        uploader_homepage = unicode(request.params.get('uploader-homepage', '')) or None
        uploader_publish = request.params.get('uploader-publish', "") == "yes"
        
        #Validate language.
        language_code = project_details.get_best_language(language)
        if not language_code:
            return h.serve_error(
             "Invalid language code provided",
             "%(project-family)s/%(project)s does not currently accept screenshots for '%(language)s'" % {
              'project-family': project_family,
              'project': project,
              'language': language,
             }
            )
            
        #Receive file.
        screenshot = request.POST['screenshot']
        
        filetype = imghdr.what('', screenshot.file.read(4096))
        if not filetype == config['quickshot.screenshot_type']:
            return h.serve_error(
             "Invalid screenshot uploaded",
             "Expected a screenshot of type %(expected)s; received %(received)s" % {
              'expected': config['quickshot.screenshot_type'],
              'received': filetype or 'unknown',
             }
            )
        screenshot.file.seek(0)
        
        filename = "%(name)s@%(time)i.%(extension)s" % {
         'name': screenshot.filename,
         'time': int(time.time()),
         'extension': config['quickshot.screenshot_type'],
        }
        
        #Store file.
        subpath = config['quickshot.data_state_pending']
        if screenshot.filename in [screenshot_name for (screenshot_name, filenames) in h.list_screenshots_accepted(project_details, language_code)]:
            subpath = config['quickshot.data_state_rejected']
        elif not screenshot.filename in [screenshot_details.screenshot_name for screenshot_details in project_details.active_screenshots]:
            subpath = config['quickshot.data_state_rejected']
            
        screenshot_destination = os.path.join(h.get_screenshots_base_path(project_details, language_code), subpath)
        if not os.path.isdir(screenshot_destination):
            old_umask = os.umask(0)
            try:
                os.makedirs(screenshot_destination, 0700)
            except Exception:
                return h.serve_error(
                 "Internal server error",
                 "Unable to create directory for %(project-family)s/%(project)s" % {
                  'project-family': project_family,
                  'project': project,
                 }
                )
            finally:
                os.umask(old_umask)
        screenshot_destination = os.path.join(screenshot_destination, filename)
        
        try:
            destination_file = open(screenshot_destination, 'w')
        except Exception:
            return h.serve_error(
             "Internal server error",
             "Unable to create file for %(project-family)s/%(project)s" % {
              'project-family': project_family,
              'project': project,
             }
            )
        else:
            try:
                shutil.copyfileobj(screenshot.file, destination_file)
            except Exception:
                return h.serve_error(
                 "Internal server error",
                 "Unable to write file %(project-family)s/%(project)s" % {
                  'project-family': project_family,
                  'project': project,
                 }
                )
            try:
                destination_file.close()
                screenshot.close()
            except Exception:
                pass
                
        #Update database.
        new_screenshot_upload = model.ProjectScreenshotUpload()
        new_screenshot_upload.p_short_name = unicode(project_family)
        new_screenshot_upload.p_version = unicode(project)
        new_screenshot_upload.language_code = language_code
        new_screenshot_upload.filename = unicode(filename)
        new_screenshot_upload.uploader = uploader
        new_screenshot_upload.uploader_name = uploader_name
        new_screenshot_upload.uploader_homepage = uploader_homepage
        new_screenshot_upload.uploader_publish = uploader_publish
        
        meta.Session.add(new_screenshot_upload)
        try:
            meta.Session.commit()
        except Exception:
            meta.Session.rollback()
            
        #Done.
        return h.serve_success(
         "Screenshot uploaded",
         '; '.join((
          "project: %(project-family)s",
          "version: %(project)s",
          "language: %(language)s",
          "screenshot: %(filename)s",
          "uploader: %(uploader)s",
          "uploader-name: %(uploader-name)s",
          "uploader-homepage: %(uploader-homepage)s",
          "uploader-anonymous: %(uploader-publish)s",
         )) % {
          'project-family': project_family,
          'project': project,
          'language': language_code,
          'filename': screenshot.filename,
          'uploader': uploader,
          'uploader-name': uploader_name or 'n/a',
          'uploader-homepage': uploader_homepage or 'n/a',
          'uploader-publish': (uploader_publish and 'no') or 'yes',
         }
        )
        
