"""Helper functions

Consists of functions to typically be used within templates, but also
available to Controllers. This module is available to templates as 'h'.
"""
import os
import random
import re

from openid.consumer.consumer import urinorm

from pylons import config, session, url
from pylons.controllers.util import redirect_to

import quickshot.model as model
import quickshot.model.meta as meta

def assert_owner(project_family):
    return bool(meta.Session.query(
     model.ProjectFamilyOwner
    ).filter(
     model.ProjectFamilyOwner.pf_short_name == project_family
    ).filter(
     model.ProjectFamilyOwner.openid == session.get('openid')
    ).count())
    
def check_presence(variable, name):
    if not variable or not variable.strip():
        session['problem_report'] = (
         ("Insufficient information provided", (
          "%(name)s not supplied" % {
           'name': name,
          },
         )),
        )
        session.save()
        return False
    return True
    
def find_accepted_screenshot(project_details, language, screenshot):
    accepted_path = os.path.join(get_screenshots_base_path(project_details), language, config['quickshot.data_state_accepted'], screenshot + '.' + config['quickshot.screenshot_type'])
    if os.path.isfile(accepted_path):
        return accepted_path
    return None
    
def find_latest_screenshot(project_details, language, screenshot):
    screenshots = order_screenshots(project_details, language, config['quickshot.data_state_pending'], screenshot)
    if screenshots:
        return os.path.join(get_screenshots_base_path(project_details, language), config['quickshot.data_state_pending'], screenshots[-1])
    return None
    
def find_reference_screenshot(project_details, preferred_language, screenshot):
    screenshot_filename = screenshot + '.' + config['quickshot.screenshot_type']
    
    if preferred_language:
        preferred_path = os.path.join(get_screenshots_base_path(project_details, preferred_language), config['quickshot.data_state_accepted'], screenshot_filename)
        if os.path.isfile(preferred_path):
            return preferred_path
            
    base_path = get_screenshots_base_path(project_details)
    candidates = [os.path.join(base_path, language, config['quickshot.data_state_accepted'], screenshot_filename) for language in os.listdir(base_path)]
    random.shuffle(candidates)
    for candidate in candidates:
        if os.path.isfile(candidate):
            return candidate
    return None
    
def get_project_details(project_family, project):
    return meta.Session.query(
     model.Project
    ).filter(
     model.Project.pf_short_name == project_family
    ).filter(
     model.Project.version == project
    ).first()
    
def get_project_family_details(project_family):
    return meta.Session.query(
     model.ProjectFamily
    ).filter(
     model.ProjectFamily.short_name == project_family
    ).first()
    
def get_screenshots_base_path(project_details, language=None):
    return os.path.normpath("%(base)s/%(project-family)s/%(project)s%(language)s" % {
     'base': config['quickshot.data_path'],
     'project-family': project_details.pf_short_name,
     'project': project_details.version,
     'language': (language and '/' + language) or '',
    })
    
def get_screenshot(project_details, screenshot):
    return meta.Session.query(
     model.ProjectScreenshot
    ).filter(
     model.ProjectScreenshot.p_short_name == project_details.pf_short_name
    ).filter(
     model.ProjectScreenshot.p_version == project_details.version
    ).filter(
     model.ProjectScreenshot.screenshot_name == screenshot
    ).first()
    
def get_screenshot_translation(screenshot_details, language):
    return meta.Session.query(
     model.ProjectScreenshotTranslation
    ).filter(
     model.ProjectScreenshotTranslation.ps_short_name == screenshot_details.p_short_name
    ).filter(
     model.ProjectScreenshotTranslation.ps_version == screenshot_details.p_version
    ).filter(
     model.ProjectScreenshotTranslation.ps_screenshot_name == screenshot_details.screenshot_name
    ).filter(
     model.ProjectScreenshotTranslation.language_code == language
    ).first()
    
def last_visited_page():
    redirect_to(session.get('last-visited-url', url('/')))
    
def list_screenshots_accepted(project_details, language):
    return _scan_filesystem(
     project_details, language,
     config['quickshot.data_state_accepted'],
     re.compile(r"^(?P<name>.+?)\." + config['quickshot.screenshot_type'] + "$")
    )
    
def list_screenshots_pending(project_details, language):
    names_accepted = set([name for (name, filenames) in list_screenshots_accepted(project_details, language)])
    pending = _scan_filesystem(
     project_details, language,
     config['quickshot.data_state_pending'],
     re.compile(r"^(?P<name>.+?)@\d+\." + config['quickshot.screenshot_type'] + "$")
    )
    
    return [(name, filenames) for (name, filenames) in pending if not name in names_accepted]
    
def normalise_openid(openid):
    if not isinstance(openid, basestring):
        raise ValueError("OpenID of type '%(type)s' cannot be processed" % {
         'type': type(openid),
        })
        
    if not openid.lower().startswith(('http://', 'https://')):
        openid = 'http://' + openid
    return urinorm.urinorm(openid)
    
def order_screenshots(project_details, language, subpath, screenshot):
    path = os.path.join(get_screenshots_base_path(project_details, language), subpath)
    regexp = re.compile("^" + screenshot + r"@(?P<timestamp>\d+)\." + config['quickshot.screenshot_type'] + "$")
    
    timestamps = []
    if os.path.isdir(path):
        for filename in os.listdir(path):
            match = regexp.match(filename)
            if match:
                timestamps.append(int(match.group('timestamp')))
    return ["%(name)s@%(time)s.%(extension)s" % {
     'name': screenshot,
     'time': timestamp,
     'extension': config['quickshot.screenshot_type'],
    } for timestamp in sorted(timestamps)]
    
def pluralise(count, singular='', plural='s'):
    return not count == 1 and plural or singular
    
def screen_error(error):
    if session.get('openid', None) == config['admin.openid']:
        return str(error)
    return "Cause unknown"
    
def serve_error(summary, details):
    return _serve_result("error", summary, details)
    
def serve_file(file_object):
    file_object.seek(0)
    chunk = file_object.read(4096) #4K chunks.
    while chunk:
        yield chunk
        chunk = file_object.read(4096)
    try:
        file_object.close()
    except Exception:
        pass
        
def serve_success(summary, details):
    return _serve_result("success", summary, details)
    
def unauthorized():
    session['problem_report'] = (
     ("You are not authorized to access the requested page", ()),
    )
    session.save()
    redirect_to(url('/'))
    
def unsupported():
    session['problem_report'] = (
     ("The requested project does not support the options you have requested", ()),
    )
    session.save()
    redirect_to(url('/'))
    
def _scan_filesystem(project_details, language, subpath, regexp):
    """
    Scans the specified directory for uploaded screenshots.
    
    The list of discovered screenshot filenames is grouped under their common
    screenshot name. The set is then reduced to exclude uninteresting
    screenshots and ordered appropriately.
    
    @type project_details: model.Project
    @param project_details: Details of the project for which screenshots are
        being scanned.
    @type language: basestr
    @param language: The language for which screenshots are being scanned.
    @type subpath: basestr
    @param subpath: The state of the screenshots to be scanned.
    @type regexp: _sre.SRE_Pattern
    @param regexp: The expression used to parse filenames.
    
    @rtype: list
    @return: A list of tuples containing common screenshot names and lists of
        filenames.
    """
    path = os.path.join(get_screenshots_base_path(project_details, language), subpath)
    
    result = {}
    if os.path.isdir(path):
        for filename in os.listdir(path):
            match = regexp.match(filename)
            if match:
                result_slice = result.get(match.group('name'))
                if result_slice is None:
                    result_slice = result[match.group('name')] = []
                result_slice.append(filename)
                
    filtered_result = []
    for screenshot in project_details.active_screenshots:
        match = result.get(screenshot.screenshot_name)
        if match:
            filtered_result.append((screenshot.screenshot_name, match))
            
    return filtered_result
    
def _serve_result(result, summary, details):
    return '\r'.join((
     'Result: %(result)s' % {'result': result,},
     'Summary: %(summary)s' % {'summary': summary,},
     'Details: %(details)s' % {'details': details,},
    ))
    
