"""The application's model objects"""
import datetime
import re

import sqlalchemy as sa
from sqlalchemy import orm

from pylons import config

from quickshot.model import meta

def init_model(engine):
    """Call me before using any of the tables or classes in the model"""
    sm = orm.sessionmaker(autoflush=True, autocommit=False, bind=engine)
    meta.engine = engine
    meta.Session = orm.scoped_session(sm)
    
#Tables needed for a Quickshot server
####################################################################################################
_project_families = sa.Table('project_families', meta.metadata,
 sa.Column('short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('friendly_name', sa.types.Unicode(255), nullable=False), #'Ubuntu Manual Project'
 sa.Column('homepage', sa.types.Unicode(255), nullable=True), #'http://ubuntu-manual.org/'
 sa.Column('public', sa.types.Boolean(), nullable=False, default=False), #True to make this family visible to everyone.
 sa.Column('active', sa.types.Boolean(), nullable=False, default=True), #False to disable all children.
) #: Holds high-level summaries of all managed project families.
class ProjectFamily(object):
    @property
    def local_languages(self):
        return set([language.language_code for language in self.languages])
orm.mapper(ProjectFamily, _project_families, order_by=('friendly_name', 'short_name'))

_projects = sa.Table('projects', meta.metadata,
 sa.Column('pf_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('version', sa.types.Unicode(255), primary_key=True), #'10.04'
 sa.Column('friendly_name', sa.types.Unicode(255), nullable=False), #'Getting Started with Ubuntu 10.04'
 sa.Column('homepage', sa.types.Unicode(255), nullable=True), #'http://ubuntu-manual.org/'
 sa.Column('resolution_x', sa.types.SmallInteger(), nullable=False), #The X-resolution for this project; <= 0 means "any".
 sa.Column('resolution_y', sa.types.SmallInteger(), nullable=False), #The Y-resolution for this project; <= 0 means "any".
 sa.Column('username', sa.types.Unicode(255), nullable=True), #The username that must be used for this project's screenshots
 sa.Column('display_name', sa.types.Unicode(255), nullable=True), #The display name that must be used for this project's screenshots
 sa.Column('expiration_date', sa.types.Date, nullable=True), #The date after which screenshot uploads will be rejected; null to disable
 sa.Column('public', sa.types.Boolean(), nullable=False, default=False), #True to make this project visible to everyone.
 sa.Column('active', sa.types.Boolean(), nullable=False, default=True), #False to disable upload/management access.
 
 sa.ForeignKeyConstraint(
  ('pf_short_name',),
  ('project_families.short_name',)
 ),
) #: Provides for versioning within project families.
class Project(object):
    @property
    def active_screenshots(self):
        return [screenshot for screenshot in self.screenshots if screenshot.active]
        
    @property
    def all_languages(self):
        return self.project_family.local_languages.union(self.local_languages)
        
    def get_best_language(self, target_language):
        if not target_language or not re.match(config['path.valid_language'], target_language):
            return None
            
        all_languages = self.all_languages
        
        if target_language in all_languages:
            return target_language
        elif target_language[:2] in all_languages:
            return target_language[:2]
        return None
        
    @property
    def is_active(self):
        if self.active and self.project_family.active:
            if self.expiration_date and self.expiration_date < datetime.datetime.utcnow().date():
                return False
        else:
            return False
        return True
        
    @property
    def is_public(self):
        return self.public and self.project_family.public
        
    @property
    def local_languages(self):
        return set([language.language_code for language in self.languages])
orm.mapper(Project, _projects,
 order_by=('pf_short_name', 'friendly_name', 'version'),
 properties={
  'project_family': orm.relation(ProjectFamily, backref=orm.backref('projects', order_by=('friendly_name', 'version'))),
 }
)

_project_family_languages = sa.Table('project_family_languages', meta.metadata,
 sa.Column('pf_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('language_code', sa.types.Unicode(5), primary_key=True), #'fr_CA'
 
 sa.ForeignKeyConstraint(
  ('pf_short_name',),
  ('project_families.short_name',),
  ondelete="CASCADE"
 ),
) #: Provides a list of languages common to all members of a project family.
class ProjectFamilyLanguage(object):
    pass
orm.mapper(ProjectFamilyLanguage, _project_family_languages,
 order_by=('pf_short_name', 'language_code'),
 properties={
  'project_family': orm.relation(ProjectFamily, backref=orm.backref('languages', order_by='language_code')),
 }
)
sa.Index('idx_project_family_languages__pf_short_name',
 ProjectFamilyLanguage.pf_short_name,
)

_project_languages = sa.Table('project_languages', meta.metadata,
 sa.Column('p_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('p_version', sa.types.Unicode(255), primary_key=True), #'10.04'
 sa.Column('language_code', sa.types.Unicode(5), primary_key=True), #'en_CA'
 
 sa.ForeignKeyConstraint(
  ('p_short_name', 'p_version',),
  ('projects.pf_short_name', 'projects.version',),
  ondelete="CASCADE"
 ),
) #: Provides a list of additional languages applicable to a specific project.
class ProjectLanguage(object):
    pass
orm.mapper(ProjectLanguage, _project_languages,
 order_by=('p_short_name', 'p_version', 'language_code'),
 properties={
  'project': orm.relation(Project, backref=orm.backref('languages', order_by='language_code')),
 }
)
sa.Index('idx_project_languages__p_short_name__p_version',
 ProjectLanguage.p_short_name,
 ProjectLanguage.p_version,
)

_project_screenshots = sa.Table('project_screenshots', meta.metadata,
 sa.Column('p_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('p_version', sa.types.Unicode(255), primary_key=True), #'10.04'
 sa.Column('screenshot_name', sa.types.Unicode(200), primary_key=True), #'login-screen'
 sa.Column('friendly_name', sa.types.Unicode(255), nullable=False), #"Ubuntu's default login screen"
 sa.Column('steps', sa.types.Text(convert_unicode=True, assert_unicode='warn'), nullable=False), #"# Turn on your computer\n= Automatic login must be disabled"
 sa.Column('display_order', sa.types.SmallInteger(), nullable=False), #The (sparse) position this screenshot will occupy when listed sequentially.
 sa.Column('command', sa.types.Text(convert_unicode=True, assert_unicode='warn'), nullable=True), #The command to have the client run, with arguments separated by linebreaks.
 sa.Column('maximize', sa.types.Boolean(), nullable=False, default=False), #Tells the client whether the window should be maximized or not.
 sa.Column('maximize_workaround', sa.types.Unicode(255), nullable=True), #Tells the client to resolve PID by process name, which is needed by things like Nautilus.
 sa.Column('fullscreen', sa.types.Boolean(), nullable=False, default=False), #Tells the client to capture the full screen, which is needed for grabbing Gnome panels.
 sa.Column('sub_rect_left', sa.types.SmallInteger(), nullable=True), #Tells the client the left edge to use for cropping.
 sa.Column('sub_rect_top', sa.types.SmallInteger(), nullable=True), #Tells the client the top edge to use for cropping.
 sa.Column('sub_rect_width', sa.types.SmallInteger(), nullable=True), #Tells the client the width to use for cropping.
 sa.Column('sub_rect_height', sa.types.SmallInteger(), nullable=True), #Tells the client the height to use for cropping.
 sa.Column('sub_rect_invert_x', sa.types.Boolean(), nullable=False, default=False), #Tells the client that cropping values relative to the left edge should be flipped.
 sa.Column('sub_rect_invert_y', sa.types.Boolean(), nullable=False, default=False), #Tells the client that cropping values relative to the top edge should be flipped.
 sa.Column('active', sa.types.Boolean(), nullable=False, default=True), #False to remove this screenshot from lists.
 
 sa.ForeignKeyConstraint(
  ('p_short_name', 'p_version',),
  ('projects.pf_short_name', 'projects.version',),
  ondelete="CASCADE"
 ),
 sa.CheckConstraint("NOT screenshot_name LIKE '%@%'"), #The only logically reserved character.
) #: Provides a list of screenshots recognized under each project, as well as base-language translation stuff.
class ProjectScreenshot(object):
    pass
orm.mapper(ProjectScreenshot, _project_screenshots,
 order_by=('p_short_name', 'p_version', 'display_order', 'screenshot_name', 'friendly_name'),
 properties={
  'project': orm.relation(Project, backref=orm.backref('screenshots', order_by=('display_order', 'screenshot_name', 'friendly_name'))),
 }
)

_project_screenshot_translations = sa.Table('project_screenshot_translations', meta.metadata,
 sa.Column('ps_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('ps_version', sa.types.Unicode(255), primary_key=True), #'10.04'
 sa.Column('ps_screenshot_name', sa.types.Unicode(200), primary_key=True), #'login-screen'
 sa.Column('language_code', sa.types.Unicode(5), primary_key=True), #'fr'
 sa.Column('friendly_name', sa.types.Unicode(255), nullable=False), #"Ubuntu's default login screen"
 sa.Column('steps', sa.types.Text(convert_unicode=True, assert_unicode='warn'), nullable=False), #"# Allumez votre ordinateur\n= N'utilisez pas la connexion automatique"
 
 sa.ForeignKeyConstraint(
  ('ps_short_name', 'ps_version', 'ps_screenshot_name'),
  ('project_screenshots.p_short_name', 'project_screenshots.p_version', 'project_screenshots.screenshot_name'),
  ondelete="CASCADE"
 ),
) #: Provides a means of offering translated screenshot instruction data to clients.
class ProjectScreenshotTranslation(object):
    pass
orm.mapper(ProjectScreenshotTranslation, _project_screenshot_translations,
 order_by=('ps_short_name', 'ps_version', 'language_code'),
 properties={
  'screenshot': orm.relation(ProjectScreenshot, backref=orm.backref('translations', order_by='language_code')),
 }
)

_project_screenshot_uploads = sa.Table('project_screenshot_uploads', meta.metadata,
 sa.Column('p_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('p_version', sa.types.Unicode(255), primary_key=True), #'10.04'
 sa.Column('language_code', sa.types.Unicode(5), primary_key=True), #'fr_FR'
 sa.Column('filename', sa.types.Unicode(255), primary_key=True), #'login-screen@1274043020.png'
 sa.Column('uploader', sa.types.Unicode(255), nullable=False), #'neil.tallim@uguu.ca'
 sa.Column('uploader_name', sa.types.Unicode(255), nullable=True), #'Neil Tallim'
 sa.Column('uploader_homepage', sa.types.Unicode(255), nullable=True), #'http://uguu.ca/'
 sa.Column('uploader_publish', sa.types.Boolean(), nullable=False, default=False), #False if the uploader wants to be anonymous.
 
 sa.ForeignKeyConstraint(
  ('p_short_name', 'p_version',),
  ('projects.pf_short_name', 'projects.version',),
  ondelete="CASCADE"
 ),
) #: Provides a means of keeping track of who uploaded which file. Entries will be removed on approval/deletion.
class ProjectScreenshotUpload(object):
    pass
orm.mapper(ProjectScreenshotUpload, _project_screenshot_uploads,
 order_by=('p_short_name', 'p_version', 'filename'),
 properties={
  'project': orm.relation(Project, backref=orm.backref('screenshot_uploads', order_by='filename')),
 }
)

_project_screenshot_approvals = sa.Table('project_screenshot_approvals', meta.metadata,
 sa.Column('p_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('p_version', sa.types.Unicode(255), primary_key=True), #'10.04'
 sa.Column('uploader', sa.types.Unicode(255), primary_key=True), #'neil.tallim@uguu.ca'
 sa.Column('uploader_name', sa.types.Unicode(255), nullable=True), #'Neil Tallim'
 sa.Column('uploader_homepage', sa.types.Unicode(255), nullable=True), #'http://uguu.ca/'
 sa.Column('uploader_publish', sa.types.Boolean(), nullable=False, default=False), #False if the uploader doesn't want to be credited.
 sa.Column('submission_count', sa.types.Integer(), nullable=False, default=0), #The number of approved screenshots related to this project.
 
 sa.ForeignKeyConstraint(
  ('p_short_name', 'p_version',),
  ('projects.pf_short_name', 'projects.version',),
  ondelete="CASCADE"
 ),
) #: Provides a means of crediting contributers.
class ProjectScreenshotApproval(object):
    pass
orm.mapper(ProjectScreenshotApproval, _project_screenshot_approvals,
 order_by=('p_short_name', 'p_version', 'submission_count DESC', 'uploader'),
 properties={
  'project': orm.relation(Project, backref=orm.backref('contributors', order_by=('submission_count DESC', 'uploader'))),
 }
)

_project_family_owners = sa.Table('project_family_owners', meta.metadata,
 sa.Column('pf_short_name', sa.types.Unicode(255), primary_key=True), #'ump'
 sa.Column('openid', sa.types.Unicode(255), primary_key=True), #'http://uguu.ca/'
 
 sa.ForeignKeyConstraint(
  ('pf_short_name',),
  ('project_families.short_name',),
  ondelete="CASCADE"
 ),
) #: Provides a list of users who are permitted to operate on a project family.
class ProjectFamilyOwner(object):
    pass
orm.mapper(ProjectFamilyOwner, _project_family_owners,
 order_by=('pf_short_name', 'openid'),
 properties={
  'project_family': orm.relation(ProjectFamily, backref=orm.backref('owners', order_by='openid')),
 }
)
sa.Index('idx_project_family_owners__pf_short_name',
 ProjectFamilyOwner.pf_short_name,
)

